//    Copyright (C) 2010  Luis Alfonso Arce González, ajaest[@]gmail.com
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

package net.ajaest.jdk.data.dict.query;

import java.util.HashSet;
import java.util.Set;

import net.ajaest.jdk.data.auxi.SimilarKanjiStrokeDemuxer;
import net.ajaest.jdk.data.auxi.database.ConnectorEnum;
import net.ajaest.jdk.data.auxi.database.QAbout;
import net.ajaest.jdk.data.auxi.database.QueryCaseEnum;
import net.ajaest.jdk.data.dict.query.kanjiFields.ClassicNelsonQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.ClassicRadicalQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.CommonStrokeMiscountQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.DeRooQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.DicNameQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.DicReferenceQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.FourCornerQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.FrequencyQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.GradeQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.JISCharsetQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.JISCodeQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.JLPTLevelQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.KanjiGraphQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.MeaningLanguageQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.MeaningQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.ReadingQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.ReadingTypeQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.SkipQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.SpahnHadamitzkyQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.StrokeCountQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.UnicodeValueQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.VariantReferenceQAbout;
import net.ajaest.jdk.data.dict.query.kanjiFields.VariantTypeQAbout;
import net.ajaest.jdk.data.kanji.KanjiGraph;
import net.ajaest.lib.utils.SystemEnums.ISO639ー1;

/**
 * This single-instance class provide methods to adapt a kanji expression in
 * order to expand or limit the range of defined kanji of the expression.
 * Example usage could be:<br>
 * 
 * &nbsp;{@code KanjiExpression ke = QueryAdaptor.limitToJôyô().meaningsInLanguaje(ISO639ー1.EN).adapt(kanjiExpression);}
 * 
 * @author Luis Alfonso Arce González
 * 
 */
public class QueryAdaptor {

	private static boolean DEMUX = false;
	private static boolean SELECT_MEANING_LANGUAJE = false;
	private static boolean LIMIT_TO_JÔYÔ_KANJI = false;
	private static boolean DEMUX_READINGS = false;
	private static boolean DEMUX_DIC_REFS = false;

	private static Integer DEMUXES;
	private static SimilarKanjiStrokeDemuxer GRAPHDEMUXER;

	private static ISO639ー1 MEANINGLANG;

	private boolean limitToJöyöKanji;
	private boolean demux;
	private boolean selectMeaningLanguaje;
	private boolean demuxReadings;
	@SuppressWarnings("unused")
	private boolean demuxDicRefs;

	private Integer demuxes;
	private SimilarKanjiStrokeDemuxer graphDemuxer;

	private ISO639ー1 meaningLang;

	private QueryAdaptor(SimilarKanjiStrokeDemuxer graphDemuxer, boolean demux, Integer demuxes,
						 boolean selectMeaningLanguaje, ISO639ー1 meaningLang, 
						 boolean limitToJöyöKanji, boolean demuxReadings, boolean demuxDicRefs) {

		this.demux = demux;
		this.selectMeaningLanguaje = selectMeaningLanguaje;
		this.limitToJöyöKanji = limitToJöyöKanji;
		this.demuxReadings = demuxReadings;
		this.demuxDicRefs = demuxDicRefs;

		this.graphDemuxer = graphDemuxer;
		this.demuxes = demuxes;

		this.meaningLang = meaningLang;

		DEMUX = demux;
		SELECT_MEANING_LANGUAJE = selectMeaningLanguaje;
		LIMIT_TO_JÔYÔ_KANJI = limitToJöyöKanji;
		DEMUX_READINGS = demuxReadings;
		DEMUX_DIC_REFS = demuxDicRefs;

		GRAPHDEMUXER = graphDemuxer;
		DEMUXES = demuxes;

		MEANINGLANG = meaningLang;
	}

	/**
	 * Sets that in the query to be adapted all occurrences of graphs will be
	 * demuxed the in the specified number of new graphs using the specified
	 * stroke demuxer
	 * 
	 * @param sksd
	 *            the stroke demuxer to be used to demux all graph occurrences
	 *            in the expression
	 * @param demuxes
	 *            count of desired graph demuxes per graph instance in query
	 * @return a query adaptor with option "demuxGraph" set on
	 */
	public static QueryAdaptor demuxGraphs(SimilarKanjiStrokeDemuxer sksd, Integer demuxes) {

		return new QueryAdaptor(sksd, true, demuxes, SELECT_MEANING_LANGUAJE, MEANINGLANG, LIMIT_TO_JÔYÔ_KANJI, DEMUX_READINGS, DEMUX_DIC_REFS);
	}

	/**
	 * Sets that in the query to be adapted all occurrences of meaning consult
	 * will be limited to an specified language
	 * 
	 * @param lang
	 *            the ISO-639-1 language in which all meaning consults will be
	 *            limited
	 * 
	 * @return a query adaptor with option "meaningsInLanguaje" set on
	 */
	public static QueryAdaptor meaningsInLanguaje(ISO639ー1 lang) {

		return new QueryAdaptor(GRAPHDEMUXER, DEMUX, DEMUXES, true, lang, LIMIT_TO_JÔYÔ_KANJI, DEMUX_READINGS, DEMUX_DIC_REFS);
	}

	/**
	 * Sets that the query to be adapted will only refer to kanjis which grade
	 * is lower than 10
	 * 
	 * @return a query adaptor with option "limitToJôyô" set on
	 */
	public static QueryAdaptor limitToJôyô() {

		return new QueryAdaptor(GRAPHDEMUXER, DEMUX, DEMUXES, SELECT_MEANING_LANGUAJE, MEANINGLANG, true, DEMUX_READINGS, DEMUX_DIC_REFS);
	}

	/**
	 * Sets that in the query to be adapted all reading consults will be demuxed
	 * in order to being able to find readings with kanji markers. For example,
	 * the reading 「たべる」 will be demuxed as follows:<br>
	 * &nbsp;&nbsp;&nbsp;&nbsp; 「たべる」<br>
	 * &nbsp;&nbsp;&nbsp;&nbsp; 「た.べる」<br>
	 * &nbsp;&nbsp;&nbsp;&nbsp; 「たべ.る」<br>
	 * &nbsp;&nbsp;&nbsp;&nbsp; 「-たべる」<br>
	 * &nbsp;&nbsp;&nbsp;&nbsp; 「たべる-」<br>
	 * 
	 * @return a query adaptor with option "demuxReadings" set on
	 */
	public static QueryAdaptor demuxReadings() {

		return new QueryAdaptor(GRAPHDEMUXER, DEMUX, DEMUXES, SELECT_MEANING_LANGUAJE, MEANINGLANG, LIMIT_TO_JÔYÔ_KANJI, true, DEMUX_DIC_REFS);
	}
	
	/* Deprecated. Now references are Integer
	/**
	 * Sets that the query to be adapted will be analyzed in order to adapt the
	 * dictionary references to the names present in the same AND clause. For
	 * example, the code:<br>
	 * <br>{@code new KanjiQuery().dictionary_name().equal("kanji_basic_book").
	 * AND().dictionary_reference().lessThan("10");}"<br>
	 * <br>
	 * will return the kanji whose references are from 001 to 099 because of the
	 * lexicographically implementation of {@code String}'s {@code compareTo}
	 * method. The right way to make this query would be: <br>
	 * <br>{@code new KanjiQuery().dictionary_name().equal("kanji_basic_book").
	 * AND().dictionary_reference().lessThan("010");}"<br>
	 * <br>
	 * However, because is not expected that the user knows the format of every
	 * dictionary's format, in most cases using this function can expand
	 * expressions in order to make the api easier to use. Adapting the first
	 * example's expression with demuxDicRefs option enabled will translate it
	 * into second example's expression.
	 * 
	 * @return a query adaptor with option "demuxDicRefs" set on
	 * /
	public static QueryAdaptor demuxDicRefs() {

		return new QueryAdaptor(GRAPHDEMUXER, DEMUX, DEMUXES, SELECT_MEANING_LANGUAJE, MEANINGLANG, LIMIT_TO_JÔYÔ_KANJI, DEMUX_READINGS, true);
	}*/

	public KanjiExpression adapt(KanjiExpression ke) {

		KanjiExpression adaptedKe = null;
		KanjiQuery kq;

		/*
		Set<String> dicNames = new HashSet<String>();
		Set<String> dicRefs = new HashSet<String>();
		StringValueQAbout strQa;
		ConnectorEnum nextConnector;*/

		KanjiExpression subKe;
		QAbout domain;
		for (int i = 0; i < ke.getDomains().size(); i++) {
			domain = ke.getDomains().get(i);
			

			kq = selectConnector(adaptedKe, domain.getBooleanCases().get(i));

			if (domain instanceof KanjiExpression) {
				subKe = adapt((KanjiExpression) domain);
				adaptedKe = addDomain(kq, subKe);
			} else if (demux && (domain instanceof KanjiGraphQAbout)) {
				subKe = demuxGraph((KanjiGraphQAbout) domain);
				if (subKe != null)
					adaptedKe = addDomain(kq, subKe);
				else
					adaptedKe = addDomain(kq, domain);
			} else if (selectMeaningLanguaje && (domain instanceof MeaningQAbout)) {
				kq = kq.meaning_language().equal(meaningLang.toString().toLowerCase()).AND();
				adaptedKe = addDomain(kq, domain);
			} else if (selectMeaningLanguaje && (domain instanceof MeaningLanguageQAbout)) {
				adaptedKe = selectCase(kq.meaning_language(), ((MeaningLanguageQAbout) domain).getQueryCase(), meaningLang.toString().toLowerCase());
			} else if (demuxReadings && (domain instanceof ReadingQAbout)) {
				subKe = demuxReadings((ReadingQAbout) domain);
				adaptedKe = addDomain(kq, subKe);
			} /*else if (demuxDicRefs && ((domain instanceof DicNameQAbout) || domain instanceof DicReferenceQAbout)) {

				if(domain.getBooleanCases().size()!=(i+1))
					nextConnector = domain.getBooleanCases().get(i+1);
				else
					nextConnector =  null;
				
				strQa = (StringValueQAbout) domain;
				if (domain instanceof DicNameQAbout)
					dicNames.add(strQa.getValue());
				else
					dicRefs.add(strQa.getValue());
				
				if ((nextConnector == null) || (nextConnector == ConnectorEnum.OR)) {
					subKe = demuxDicRefs(dicNames, dicRefs);
					adaptedKe = addDomain(kq, subKe);
					dicNames = new HashSet<String>();
					dicRefs = new HashSet<String>();
				}
			} */else {
				adaptedKe = addDomain(kq, domain);
			}
		}

		if (limitToJöyöKanji) {
			adaptedKe = new KanjiQuery().expression(adaptedKe).AND().grade().equalsOrLessThan(10);
		}

		return adaptedKe;
	}

	/*
	private KanjiExpression demuxDicRefs(Set<String> dicNames, Set<String> dicRefs) {


		return null;
	}
	*/
	private KanjiExpression demuxReadings(ReadingQAbout domain) {
		
		Set<String> strs = new HashSet<String>();

		String mainReading = domain.getValue();

		// okurigana markers

		String demuxed;
		for (int i = 1; i < mainReading.length(); i++) {
			demuxed = mainReading.substring(0, i);
			demuxed += '.';
			demuxed += mainReading.substring(i);
			strs.add(demuxed);
		}

		// prefixes and suffixes
		Set<String> addTostrs = new HashSet<String>();
		for (String dmx1 : strs) {
			addTostrs.add(dmx1 + "-");
			addTostrs.add("-" + dmx1);
		}

		strs.addAll(addTostrs);

		strs.add(mainReading + "-");
		strs.add("-" + mainReading);

		// builds expression

		KanjiExpression ke = new KanjiQuery().reading().equal(mainReading);

		for (String s : strs) {
			ke = ke.OR().reading().equal(s);
		}

		return ke;
	}

	private KanjiExpression demuxGraph(KanjiGraphQAbout kgqa) {

		Set<KanjiGraph> demuxed = graphDemuxer.getSimilar(kgqa.getValue(), demuxes);
		
		KanjiExpression subKe = new KanjiQuery().kanji_graph().equal(kgqa.getValue());
		KanjiQuery kq = subKe.OR();
		for(KanjiGraph kg : demuxed){
			subKe = kq.kanji_graph().equal(kg);
			kq = subKe.OR();
		}

		return subKe;
	}
	
	private KanjiQuery selectConnector(KanjiExpression ke, ConnectorEnum connector) {

		switch (connector) {
			case AND :
				return ke.AND();
			case OR :
				return ke.OR();
			case FIRST :
				return new KanjiQuery();
			default :
				throw new IllegalArgumentException();
		}
	}

	private KanjiExpression addDomain(KanjiQuery kq, QAbout domain) {

		KanjiExpression ke;

		if (domain instanceof KanjiExpression) {
			ke = kq.expression((KanjiExpression) domain);
		} else if (domain instanceof StringValueQAbout) {

			StringValueQAbout strDomain = (StringValueQAbout) domain;

			if (domain instanceof DeRooQAbout) {
				ke = selectCase(kq.De_Roo_code(), strDomain.getQueryCase(), strDomain.getValue());
			} else if (domain instanceof DicNameQAbout) {
				ke = selectCase(kq.dictionary_name(), strDomain.getQueryCase(), strDomain.getValue());
			} else if (domain instanceof FourCornerQAbout) {
				ke = selectCase(kq.four_corner_code(), strDomain.getQueryCase(), strDomain.getValue());
			} else if (domain instanceof JISCharsetQAbout) {
				ke = selectCase(kq.JIS_charset(), strDomain.getQueryCase(), strDomain.getValue());
			} else if (domain instanceof JISCodeQAbout) {
				ke = selectCase(kq.JIS_code_value(), strDomain.getQueryCase(), strDomain.getValue());
			} else if (domain instanceof MeaningLanguageQAbout) {
				ke = selectCase(kq.meaning_language(), strDomain.getQueryCase(), strDomain.getValue());
			} else if (domain instanceof MeaningQAbout) {
				ke = selectCase(kq.meaning(), strDomain.getQueryCase(), strDomain.getValue());
			} else if (domain instanceof ReadingTypeQAbout) {
				ke = selectCase(kq.reading_type(), strDomain.getQueryCase(), strDomain.getValue());
			} else if (domain instanceof ReadingQAbout) {
				ke = selectCase(kq.reading(), strDomain.getQueryCase(), strDomain.getValue());
			} else if (domain instanceof SkipQAbout) {
				ke = selectCase(kq.skip_code(), strDomain.getQueryCase(), strDomain.getValue());
			} else if (domain instanceof SpahnHadamitzkyQAbout) {
				ke = selectCase(kq.SpahnーHadamizky_code(), strDomain.getQueryCase(), strDomain.getValue());
			} else if (domain instanceof VariantReferenceQAbout) {
				ke = selectCase(kq.variant_reference(), strDomain.getQueryCase(), strDomain.getValue());
			} else if (domain instanceof VariantTypeQAbout) {
				ke = selectCase(kq.variant_type(), strDomain.getQueryCase(), strDomain.getValue());
			} else {
				throw new IllegalArgumentException();
			}


		} else if (domain instanceof IntegerValueQAbout) {

			IntegerValueQAbout intDomain = (IntegerValueQAbout) domain;

			if (domain instanceof ClassicNelsonQAbout) {
				ke = selectCase(kq.classic_nelson(), intDomain.getQueryCase(), intDomain.getValue());
			} else if (domain instanceof DicReferenceQAbout) {
				ke = selectCase(kq.dictionary_reference(), intDomain.getQueryCase(), intDomain.getValue());
			} else if (domain instanceof ClassicRadicalQAbout) {
				ke = selectCase(kq.classic_radical(), intDomain.getQueryCase(), intDomain.getValue());
			} else if (domain instanceof CommonStrokeMiscountQAbout) {
				ke = selectCase(kq.stroke_miscount(), intDomain.getQueryCase(), intDomain.getValue());
			} else if (domain instanceof FrequencyQAbout) {
				ke = selectCase(kq.frequency(), intDomain.getQueryCase(), intDomain.getValue());
			} else if (domain instanceof GradeQAbout) {
				ke = selectCase(kq.grade(), intDomain.getQueryCase(), intDomain.getValue());
			} else if (domain instanceof JLPTLevelQAbout) {
				ke = selectCase(kq.JLPT_level(), intDomain.getQueryCase(), intDomain.getValue());
			} else if (domain instanceof StrokeCountQAbout) {
				ke = selectCase(kq.stroke_count(), intDomain.getQueryCase(), intDomain.getValue());
			} else if (domain instanceof UnicodeValueQAbout) {
				ke = selectCase(kq.unicode_value(), intDomain.getQueryCase(), intDomain.getValue());
			} else {
				throw new IllegalArgumentException();
			}

		} else {
			if (domain instanceof ValueQAbout<?>) {
				
				Object unknowValue = ((ValueQAbout<?>) domain).getValue();
				
				if(unknowValue instanceof KanjiGraph){
					ke = selectCase(kq.kanji_graph(), ((ValueQAbout<?>) domain).getQueryCase(), (KanjiGraph) unknowValue);
				} else {
					throw new IllegalArgumentException();
				}
			} else {
				throw new IllegalArgumentException();
			}

		}

		return ke;
	}

	private <E> KanjiExpression selectCase(ValueQAbout<E> vqa, QueryCaseEnum caze, E value) {

		switch (caze) {
			case EQUALS:
				return vqa.equal(value);
			case GREATER_OR_EQUALS_THAN:
				return vqa.equaslOrGreatherThan(value);
			case GREATER_THAN:
				return vqa.greatherThan(value);
			case LESS_OR_EQUALS_THAN:
				return vqa.equalsOrLessThan(value);
			case LESS_THAN:
				return vqa.lessThan(value);
			case NOT_EQUALS :
				return vqa.notEquals(value);
			case NULL :
				return vqa.isNull();
			default :
				throw new IllegalArgumentException();
				
		}
	}

	@Override
	protected void finalize() {
		DEMUX = false;
		SELECT_MEANING_LANGUAJE = false;
		LIMIT_TO_JÔYÔ_KANJI = false;
		DEMUX_READINGS = false;
	}
}
